import numpy as np
from numba import njit

NAN = np.float64(np.nan)


@njit(cache=True)
def _ifd_acm(P, Q, delta, dist_type):
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    P_subedges_num = _steiner_subedges(P, delta)
    Q_subedges_num = _steiner_subedges(Q, delta)

    p, q = len(P), len(Q)
    P_idxs = np.empty(p, dtype=np.int_)
    if p > 0:
        P_idxs[0] = 0
        P_idxs[1:] = np.cumsum(P_subedges_num)
    Q_idxs = np.empty(q, dtype=np.int_)
    if q > 0:
        Q_idxs[0] = 0
        Q_idxs[1:] = np.cumsum(Q_subedges_num)

    pp = 0 if p == 0 else np.sum(P_subedges_num) + 1
    qq = 0 if q == 0 else np.sum(Q_subedges_num) + 1
    B = np.empty((pp, q), dtype=np.float64)
    if pp > 0 and q > 0:
        B[0, 0] = 0
        B[-1, -1] = NAN
    L = np.empty((p, qq), dtype=np.float64)
    if p > 0 and qq > 0:
        L[0, 0] = 0
        L[-1, -1] = NAN

    for i in range(p - 1):
        p_pts = _steiner_pts(P[i : i + 2], P_subedges_num[i])
        for j in range(q - 1):
            q_pts = _steiner_pts(Q[j : j + 2], Q_subedges_num[j])

            p_costs = B[P_idxs[i] : P_idxs[i + 1] + 1, j]
            q_costs = L[i, Q_idxs[j] : Q_idxs[j + 1] + 1]

            p1, q1 = _update_cell(
                p_pts,
                q_pts,
                p_costs,
                B[P_idxs[i] : P_idxs[i + 1] + 1, j + 1],
                q_costs,
                L[i + 1, Q_idxs[j] : Q_idxs[j + 1] + 1],
                i == 0,
                j == 0,
                i == p - 2,
                j == q - 2,
                dist_type,
            )

            p_costs[-1:] = p1
            q_costs[-1:] = q1

    return B, L


@njit(cache=True)
def _ifd_acm_1d(P, Q, delta, dist_type):
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    P_subedges_num = _steiner_subedges(P, delta)
    Q_subedges_num = _steiner_subedges(Q, delta)

    p, q = len(P), len(Q)
    P_idxs = np.empty(p, dtype=np.int_)
    if p > 0:
        P_idxs[0] = 0
        P_idxs[1:] = np.cumsum(P_subedges_num)
    Q_idxs = np.empty(q, dtype=np.int_)
    if q > 0:
        Q_idxs[0] = 0
        Q_idxs[1:] = np.cumsum(Q_subedges_num)

    pp = 0 if p == 0 else np.sum(P_subedges_num) + 1
    qq = 0 if q == 0 else np.sum(Q_subedges_num) + 1
    B = np.empty(pp, dtype=np.float64)
    if pp > 0:
        B[0] = 0
        B[-1] = NAN
    L = np.empty(qq, dtype=np.float64)
    if qq > 0:
        L[0] = 0
        L[-1] = NAN

    # TODO: parallelize this i-loop.
    # Must ensure that cell (i - 1, j) is computed before (i, j).
    p0 = B[:1]  # will be updated during i-loop when j == 0
    for i in range(p - 1):
        p_pts = _steiner_pts(P[i : i + 2], P_subedges_num[i])
        q0 = L[:1]  # will be updated during j-loop
        for j in range(q - 1):
            q_pts = _steiner_pts(Q[j : j + 2], Q_subedges_num[j])

            if j == 0:
                p_costs = np.concatenate((p0, B[P_idxs[i] + 1 : P_idxs[i + 1] + 1]))
            else:
                p_costs = B[P_idxs[i] : P_idxs[i + 1] + 1].copy()
            q_costs = np.concatenate((q0, L[Q_idxs[j] + 1 : Q_idxs[j + 1] + 1]))

            p1, q1 = _update_cell(
                p_pts,
                q_pts,
                p_costs,
                B[P_idxs[i] : P_idxs[i + 1] + 1],
                q_costs,
                L[Q_idxs[j] : Q_idxs[j + 1] + 1],
                i == 0,
                j == 0,
                i == p - 2,
                j == q - 2,
                dist_type,
            )

            # store for the next loops
            if j == 0:
                p0 = p1
            q0 = q1

    return B, L


@njit(cache=True)
def _steiner_subedges(vert, delta):
    N, _ = vert.shape
    if N == 0:
        N = 1
    edge_lens = np.empty(N - 1, dtype=np.float64)
    for i in range(N - 1):
        edge_lens[i] = np.linalg.norm(vert[i + 1] - vert[i])
    subedges_num = np.ceil(edge_lens / delta).astype(np.int_)
    return subedges_num


@njit(cache=True)
def _steiner_pts(P1P2, n):
    _, D = P1P2.shape
    pts = np.empty((n + 1, D), dtype=np.float64)
    v = P1P2[1] - P1P2[0]
    for i in range(n):
        pts[i] = P1P2[0] + (i / n) * v
    pts[n] = P1P2[1]
    return pts


@njit(cache=True)
def _update_cell(
    p_pts,
    q_pts,
    p_costs,
    p_costs_out,
    q_costs,
    q_costs_out,
    p_is_initial,
    q_is_initial,
    p_is_last,
    q_is_last,
    dist_type,
):
    P1, Q1 = p_pts[0], q_pts[0]
    u, v, b, delta_P, delta_Q = _cell_info(p_pts, q_pts)

    # Will be reused for each border point (t) to find best starting point (s).
    p_cost_candidates = np.empty(len(p_pts), dtype=np.float64)
    q_cost_candidates = np.empty(len(q_pts), dtype=np.float64)
    s = np.empty((2,), dtype=np.float64)
    t = np.empty((2,), dtype=np.float64)

    # compute upper boundary
    t[1] = delta_Q * (len(q_pts) - 1)
    if p_is_initial:  # No steiner points on left boundary; just check [0, 0]
        q_end_idx = 1
    else:
        q_end_idx = len(q_pts)
    if q_is_last:  # No need steiner points on upper boundary. Just check corner point.
        p_start_idx = len(p_pts) - 1
    else:
        p_start_idx = 0
    for i in range(p_start_idx, len(p_pts)):  # Fill p_costs_out[i]
        t[0] = delta_P * i

        s[0] = 0
        for j in range(0, q_end_idx):
            s[1] = delta_Q * j
            cost = _cell_owc(s, t, P1, Q1, u, v, b, dist_type)
            q_cost_candidates[j] = q_costs[j] + cost

        s[1] = 0
        if q_is_initial:  # No steiner points on bottom boundary; just check [0, 0]
            p_end_idx = 1
        else:
            p_end_idx = i + 1
        p_cost_candidates[0] = q_cost_candidates[0]  # cost from [0, 0] already known.
        for i_ in range(1, p_end_idx):  # let bottom border points be (s). (to right)
            s[0] = delta_P * i_
            cost = _cell_owc(s, t, P1, Q1, u, v, b, dist_type)
            p_cost_candidates[i_] = p_costs[i_] + cost

        p_costs_out[i] = min(
            np.min(p_cost_candidates[:p_end_idx]), np.min(q_cost_candidates[:q_end_idx])
        )

    # compute right boundary
    t[0] = delta_P * (len(p_pts) - 1)
    if q_is_initial:  # No steiner points on bottom boundary; just check [0, 0]
        p_end_idx = 1
    else:
        p_end_idx = len(p_pts)
    if p_is_last:  # No need steiner points on right boundary. Just check corner point.
        q_start_idx = len(q_pts) - 1
    elif q_is_initial:
        q_start_idx = 0
    else:  # LR corner already computed by the lower cell.
        q_start_idx = 1
    # No need to compute the last j (already done by P loop just above)
    for j in range(q_start_idx, len(q_pts) - 1):
        t[1] = delta_Q * j

        s[1] = 0
        for i in range(0, p_end_idx):
            s[0] = delta_P * i
            cost = _cell_owc(s, t, P1, Q1, u, v, b, dist_type)
            p_cost_candidates[i] = p_costs[i] + cost

        s[0] = 0
        if p_is_initial:  # No steiner points on left boundary; just check [0, 0]
            q_end_idx = 1
        else:
            q_end_idx = j + 1
        q_cost_candidates[0] = p_cost_candidates[0]  # cost from [0, 0] already known.
        for j_ in range(1, q_end_idx):  # cost from [0, 0] already known.
            s[1] = delta_Q * j_
            cost = _cell_owc(s, t, P1, Q1, u, v, b, dist_type)
            q_cost_candidates[j_] = q_costs[j_] + cost

        q_costs_out[j] = min(
            np.min(p_cost_candidates[:p_end_idx]), np.min(q_cost_candidates[:q_end_idx])
        )
    q_costs_out[-1] = p_costs_out[-1]  # Fill the last j

    # Lower-right corner and upper-left corner of cells.
    return q_costs_out[:1], p_costs_out[:1]


@njit(cache=True)
def _cell_info(P_pts, Q_pts):
    P1, P2 = P_pts[0], P_pts[-1]
    Q1, Q2 = Q_pts[0], Q_pts[-1]
    P1P2 = P2 - P1
    Q1Q2 = Q2 - Q1
    L1 = np.linalg.norm(P1P2)
    L2 = np.linalg.norm(Q1Q2)

    if L1 == 0:
        u = np.array([0, 0], np.float64)
    else:
        u = (P1P2) / L1
    if L2 == 0:
        v = np.array([0, 0], np.float64)
    else:
        v = (Q1Q2) / L2

    # Find lm: y = x + b
    w = P1 - Q1
    uv = np.dot(u, v)
    if uv == 1:
        # P and Q are parallel; equations degenerate into s - (u.v)t = -u.w
        b = np.dot(u, w)
    elif uv == -1:
        # P and Q are antiparallel.
        # Any value is OK, so just set b=0.
        b = np.float64(0)
    else:
        # P and Q intersects.
        # Find points P(s) and Q(t) where P and Q intersects.
        # (s, t) is on y = x + b
        A = np.array([[1, -uv], [-uv, 1]], dtype=np.float64)
        B = np.array([-np.dot(u, w), np.dot(v, w)], dtype=np.float64)
        s, t = np.linalg.solve(A, B)
        b = t - s

    delta_P = L1 / (len(P_pts) - 1)
    delta_Q = L2 / (len(Q_pts) - 1)
    return u, v, b, delta_P, delta_Q


@njit(cache=True)
def _cell_owc(s, t, P1, Q1, u, v, b, dist_type):
    P_s = P1 + u * s[0]
    P_t = P1 + u * t[0]
    Q_s = Q1 + v * s[1]
    Q_t = Q1 + v * t[1]

    if s[1] > s[0] + b:
        cs_x, cs_y = s[1] - b, s[1]
    else:
        cs_x, cs_y = s[0], s[0] + b
    if t[1] < t[0] + b:
        ct_x, ct_y = t[1] - b, t[1]
    else:
        ct_x, ct_y = t[0], t[0] + b

    if cs_x < ct_x:  # pass through lm
        P_cs = P1 + u * cs_x
        P_ct = P1 + u * ct_x
        Q_cs = Q1 + v * cs_y
        Q_ct = Q1 + v * ct_y

        if s[1] > s[0] + b:  # right
            s_to_cs = _linepoint_cost(P_s, P_cs, Q_s, dist_type)
        else:  # up
            s_to_cs = _linepoint_cost(Q_s, Q_cs, P_s, dist_type)
        cs_to_ct = _lineline_cost(P_cs, P_ct, Q_cs, Q_ct, dist_type)
        if t[1] > t[0] + b:  # up
            ct_to_t = _linepoint_cost(Q_ct, Q_t, P_t, dist_type)
        else:  # right
            ct_to_t = _linepoint_cost(P_ct, P_t, Q_t, dist_type)
        cost = s_to_cs + cs_to_ct + ct_to_t

    else:  # pass c'
        if s[1] > s[0] + b:  # right -> up
            cost1 = _linepoint_cost(P_s, P_t, Q_s, dist_type)
            cost2 = _linepoint_cost(Q_s, Q_t, P_t, dist_type)
        else:  # up -> right
            cost1 = _linepoint_cost(Q_s, Q_t, P_s, dist_type)
            cost2 = _linepoint_cost(P_s, P_t, Q_t, dist_type)
        cost = cost1 + cost2

    return cost


@njit(cache=True)
def _cell_owp(p_n, q_n, s_idx, t_idx, delta_P, delta_Q, b):
    W = delta_P * (p_n - 1)
    H = delta_P * (q_n - 1)

    s = [s_idx[0] * delta_P, s_idx[1] * delta_Q]
    t = [t_idx[0] * delta_P, t_idx[1] * delta_Q]

    verts = np.empty((4, 2), dtype=np.float64)
    verts[0] = [s_idx[0] / (p_n - 1), s_idx[1] / (q_n - 1)]
    count = 1

    if s[1] > s[0] + b:
        cs_x, cs_y = s[1] - b, s[1]
    else:
        cs_x, cs_y = s[0], s[0] + b
    if t[1] < t[0] + b:
        ct_x, ct_y = t[1] - b, t[1]
    else:
        ct_x, ct_y = t[0], t[0] + b

    if cs_x < ct_x:  # pass through lm
        if s[1] > s[0] + b:  # right
            l0_zero = s[0] == cs_x
        else:  # up
            l0_zero = s[1] == cs_y
        l1_zero = cs_x == ct_x and cs_y == ct_y
        if t[1] > t[0] + b:  # up
            l2_zero = ct_y == t[1]
        else:  # right
            l2_zero = ct_x == t[0]

        if not l0_zero:
            verts[count] = [cs_x / W, cs_y / H]
            count += 1
        if not l1_zero:
            verts[count] = [ct_x / W, ct_y / H]
            count += 1
        if not l2_zero:
            verts[count] = [t_idx[0] / (p_n - 1), t_idx[1] / (q_n - 1)]
            count += 1

    else:  # pass c'
        if s[1] > s[0] + b:  # right -> up
            l0_zero = s_idx[0] == t_idx[0]
            l1_zero = s_idx[1] == t_idx[1]
            c_prime = [t_idx[0] / (p_n - 1), s_idx[1] / (q_n - 1)]
        else:  # up -> right
            l0_zero = s_idx[1] == t_idx[1]
            l1_zero = s_idx[0] == t_idx[0]
            c_prime = [s_idx[0] / (p_n - 1), t_idx[1] / (q_n - 1)]

        if not l0_zero:
            verts[count] = c_prime
            count += 1
        if not l1_zero:
            verts[count] = [t_idx[0] / (p_n - 1), t_idx[1] / (q_n - 1)]
            count += 1
    return verts, count


@njit(cache=True)
def _linepoint_cost(a, b, p, dist_type):
    if dist_type == "euclidean":
        # TODO: implement numerical integration
        raise NotImplementedError
    elif dist_type == "squared_euclidean":
        ab = b - a
        pa = a - p
        A = np.dot(ab, ab)
        B = 2 * np.dot(ab, pa)
        C = np.dot(pa, pa)
        ret = (A / 3 + B / 2 + C) * np.sqrt(A)
    else:
        raise ValueError("Unknown type of distance.")
    return ret


@njit(cache=True)
def _lineline_cost(a, b, c, d, dist_type):
    if dist_type == "euclidean":
        # TODO: implement numerical integration
        raise NotImplementedError
    elif dist_type == "squared_euclidean":
        u = b - a
        v = d - c
        w = a - c
        vu = u - v
        A = np.dot(vu, vu)
        B = 2 * np.dot(vu, w)
        C = np.dot(w, w)
        D = np.dot(u, u)
        E = np.dot(v, v)
        ret = (A / 3 + B / 2 + C) * (np.sqrt(D) + np.sqrt(E))
    else:
        raise ValueError("Unknown type of distance.")
    return ret


@njit(cache=True)
def _ifd_owp(P, Q, B, L, delta, dist_type):
    P_subedges_num = _steiner_subedges(P, delta)
    Q_subedges_num = _steiner_subedges(Q, delta)

    p, q = len(P), len(Q)
    P_idxs = np.empty(p, dtype=np.int_)
    if p > 0:
        P_idxs[0] = 0
        P_idxs[1:] = np.cumsum(P_subedges_num)
    Q_idxs = np.empty(q, dtype=np.int_)
    if q > 0:
        Q_idxs[0] = 0
        Q_idxs[1:] = np.cumsum(Q_subedges_num)

    # Path passes (p + q - 3) cells. Each cell has max 4 vertices.
    # (p + q - 4) vertices overlap.
    MAX_PATH_VERT_NUM = (p + q - 3) * 4 - (p + q - 4)
    PATH = np.empty((MAX_PATH_VERT_NUM, 2), dtype=np.float64)
    COUNT = 0

    if p > 0 and q > 0:
        PATH[0] = [p - 1, q - 1]
        COUNT += 1

    i, j = p - 2, q - 2  # cell indices
    t_idx = np.array((P_subedges_num[i], Q_subedges_num[j]))
    while True:
        p_pts = _steiner_pts(P[i : i + 2], P_subedges_num[i])
        q_pts = _steiner_pts(Q[j : j + 2], Q_subedges_num[j])
        p_costs = B[P_idxs[i] : P_idxs[i + 1] + 1, j]
        q_costs = L[i, Q_idxs[j] : Q_idxs[j + 1] + 1]
        s_idx, path, count = _cell_path(
            t_idx, p_pts, q_pts, p_costs, q_costs, i == 0, j == 0, dist_type
        )
        path = path[: count - 1]
        path[:, 0] += i
        path[:, 1] += j
        for i_ in range(count - 1):
            PATH[COUNT + i_] = path[count - i_ - 2]
        COUNT += count - 1

        if i <= 0 and j <= 0:
            break

        t_idx[:] = s_idx
        if s_idx[0] == 0 and i > 0:
            i -= 1
            t_idx[0] = P_subedges_num[i]
        if s_idx[1] == 0 and j > 0:
            j -= 1
            t_idx[1] = Q_subedges_num[j]

    return PATH[:COUNT]


@njit(cache=True)
def _cell_path(
    t_idx,
    p_pts,
    q_pts,
    p_costs,
    q_costs,
    p_is_initial,
    q_is_initial,
    dist_type,
):
    P1, Q1 = p_pts[0], q_pts[0]
    u, v, b, delta_P, delta_Q = _cell_info(p_pts, q_pts)

    p_cost_candidates = np.empty(len(p_pts), dtype=np.float64)
    q_cost_candidates = np.empty(len(q_pts), dtype=np.float64)
    t = np.array((delta_P * t_idx[0], delta_Q * t_idx[1]), dtype=np.float64)
    s = np.empty((2,), dtype=np.float64)  # will be updated and reused

    if q_is_initial:  # No need to check points on bottom boundary.
        p_end_idx = 0
    else:
        p_end_idx = t_idx[0]
    s[1] = 0
    for i in range(0, p_end_idx + 1):
        s[0] = delta_P * i
        cost = _cell_owc(s, t, P1, Q1, u, v, b, dist_type)
        p_cost_candidates[i] = p_costs[i] + cost

    # [0, 0] already computed
    q_cost_candidates[0] = p_cost_candidates[0]
    if p_is_initial:  # No need to check steiner points on left boundary.
        q_end_idx = 0
    else:
        q_end_idx = t_idx[1]
    s[0] = 0
    for j in range(1, q_end_idx + 1):
        s[1] = delta_Q * j
        cost = _cell_owc(s, t, P1, Q1, u, v, b, dist_type)
        q_cost_candidates[j] = q_costs[j] + cost

    p_min_idx = np.argmin(p_cost_candidates[: p_end_idx + 1])
    pmin = p_cost_candidates[p_min_idx]
    q_min_idx = np.argmin(q_cost_candidates[: q_end_idx + 1])
    qmin = q_cost_candidates[q_min_idx]
    if pmin < qmin:
        s_idx = [p_min_idx, 0]
    else:
        s_idx = [0, q_min_idx]
    path, count = _cell_owp(len(p_pts), len(q_pts), s_idx, t_idx, delta_P, delta_Q, b)

    return s_idx, path, count



import numpy as np
from numba import njit


@njit(cache=True)
def _dtw_acm(P, Q, dist_type):
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    p, q = len(P), len(Q)
    ret = np.empty((p, q), dtype=np.float64)

    if p > 0 and q > 0:
        ret[0, 0] = _dist(P[0], Q[0], dist_type)

    for i in range(1, p):
        ret[i, 0] = ret[i - 1, 0] + _dist(P[i], Q[0], dist_type)

    for j in range(1, q):
        ret[0, j] = ret[0, j - 1] + _dist(P[0], Q[j], dist_type)

    for i in range(1, p):
        for j in range(1, q):
            ret[i, j] = min(ret[i - 1, j], ret[i, j - 1], ret[i - 1, j - 1]) + _dist(
                P[i], Q[j], dist_type
            )
    return ret


@njit(cache=True)
def _dtw_acm_1d(P, Q, dist_type):
    # Equivalent to _dtw_acm, but keep 1d array instead of 2d.
    # Memory efficient, but cannot do backtracking.
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    p, q = len(P), len(Q)
    ret = np.empty(q, dtype=np.float64)

    if p > 0 and q > 0:
        ret[0] = _dist(P[0], Q[0], dist_type)

    for j in range(1, q):
        ret[j] = ret[j - 1] + _dist(P[0], Q[j], dist_type)

    for i in range(1, p):
        left = ret[0]
        ret[0] = left + np.linalg.norm(P[i] - Q[0])
        for j in range(1, q):
            diag = left
            left = ret[j]
            ret[j] = min(diag, left, ret[j - 1]) + _dist(P[i], Q[j], dist_type)

    return ret


@njit(cache=True)
def _dist(p, q, dist_type):
    if dist_type == "euclidean":
        ret = np.linalg.norm(p - q)
    elif dist_type == "squared_euclidean":
        ret = np.linalg.norm(p - q) ** 2
    else:
        raise ValueError("Unknown type of distance.")
    return ret


@njit(cache=True)
def _dtw_owp(acm):
    p, q = acm.shape
    path = np.empty((p + q - 1, 2), dtype=np.int_)
    path_len = np.int_(0)

    i, j = p - 1, q - 1
    path[path_len] = [i, j]
    path_len += np.int_(1)

    while i > 0 or j > 0:
        if i == 0:
            j -= 1
        elif j == 0:
            i -= 1
        else:
            d = min(acm[i - 1, j], acm[i, j - 1], acm[i - 1, j - 1])
            if acm[i - 1, j] == d:
                i -= 1
            elif acm[i, j - 1] == d:
                j -= 1
            else:
                i -= 1
                j -= 1

        path[path_len] = [i, j]
        path_len += np.int_(1)
    return path[-(len(path) - path_len + 1) :: -1, :]



import numpy as np

from curvesimilarities import ifd, ifd_owp


def test_ifd():
    P, Q = [[0, 0], [1, 0]], [[0, 1], [1, 1]]
    assert ifd(np.asarray(P), np.asarray(Q), 1, "squared_euclidean") == 2
    assert ifd(np.asarray(Q), np.asarray(P), 1, "squared_euclidean") == 2

    P, Q = [[0, 0], [0.5, 0], [1, 0]], [[0, 1], [1, 1]]
    assert ifd(np.asarray(P), np.asarray(Q), 0.5, "squared_euclidean") == 2
    assert ifd(np.asarray(Q), np.asarray(P), 0.5, "squared_euclidean") == 2

    P, Q = [[0, 0], [0.5, 0], [1, 0]], [[0, 1], [0.5, 1], [1, 1]]
    assert ifd(np.asarray(P), np.asarray(Q), 0.5, "squared_euclidean") == 2
    assert ifd(np.asarray(Q), np.asarray(P), 0.5, "squared_euclidean") == 2


def test_ifd_degenerate(P_vert, Q_vert):
    np.isnan(ifd(P_vert[:0], Q_vert[:0], 0.1, "squared_euclidean"))


def test_ifd_owp():
    P, Q = [[0, 0], [1, 0]], [[0, 1], [1, 1]]
    assert ifd_owp(np.asarray(P), np.asarray(Q), 1, "squared_euclidean")[0] == 2
    assert ifd_owp(np.asarray(Q), np.asarray(P), 1, "squared_euclidean")[0] == 2

    P, Q = [[0, 0], [0.5, 0], [1, 0]], [[0, 1], [1, 1]]
    assert ifd_owp(np.asarray(P), np.asarray(Q), 0.5, "squared_euclidean")[0] == 2
    assert ifd_owp(np.asarray(Q), np.asarray(P), 0.5, "squared_euclidean")[0] == 2

    P, Q = [[0, 0], [0.5, 0], [1, 0]], [[0, 1], [0.5, 1], [1, 1]]
    assert ifd_owp(np.asarray(P), np.asarray(Q), 0.5, "squared_euclidean")[0] == 2
    assert ifd_owp(np.asarray(Q), np.asarray(P), 0.5, "squared_euclidean")[0] == 2


def test_ifd_owp_failedcases():
    P = [
        [403, 250],
        [403, 253],
        [402, 254],
    ]
    Q = [
        [355.75, 243.0],
        [355.89, 244.5],
        [355.75, 246.0],
    ]
    _, owp = ifd_owp(np.asarray(P), np.asarray(Q), 5.0, "squared_euclidean")
    assert owp[-1, 0] == np.sum(np.linalg.norm(np.diff(P, axis=0), axis=-1))
    assert owp[-1, 1] == np.sum(np.linalg.norm(np.diff(Q, axis=0), axis=-1))



import numpy as np
from numba import njit

EPSILON = np.finfo(np.float64).eps
NAN = np.float64(np.nan)
INF = np.float64(np.inf)


@njit(cache=True)
def _fd(P, Q, rel_tol, abs_tol):
    """Algorithm 3 of Alt & Godau (1995)."""
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    p, q = len(P), len(Q)

    if not (p > 0 and q > 0):
        return NAN

    crit = _critical_values(P, Q)

    # binary search
    start, end = 0, len(crit) - 1
    B, L = _reachable_boundaries_1d(P, Q, crit[start])
    if B[-1, 1] == 1 or L[-1, 1] == 1:
        end = start
    while end - start > 1:
        mid = (start + end) // 2
        B, L = _reachable_boundaries_1d(P, Q, crit[mid])
        if B[-1, 1] == 1 or L[-1, 1] == 1:
            end = mid
        else:
            start = mid

    # parametric search
    e1, e2 = crit[start], crit[end]
    while e2 - e1 > max(rel_tol * e2, abs_tol):
        mid = (e1 + e2) / 2
        if (mid - e1 < EPSILON) or (e2 - mid < EPSILON):
            break
        B, L = _reachable_boundaries_1d(P, Q, mid)
        if B[-1, 1] == 1 or L[-1, 1] == 1:
            e2 = mid
        else:
            e1 = mid
    return e2


@njit(cache=True)
def _critical_values(P, Q):
    p, q = len(P), len(Q)
    MAX_A = max(np.linalg.norm(P[0] - Q[0]), np.linalg.norm(P[-1] - Q[-1]))
    crit_a = np.array((MAX_A,))

    crit_b = np.empty(p * (q - 1) + (p - 1) * q, dtype=np.float64)
    count = 0
    for i in range(p - 1):
        for j in range(q):
            dist, _ = _critical_b(P[i], P[i + 1], Q[j])
            if dist > MAX_A:
                crit_b[count] = dist
                count += 1
    for i in range(p):
        for j in range(q - 1):
            dist, _ = _critical_b(Q[j], Q[j + 1], P[i])
            if dist > MAX_A:
                crit_b[count] = dist
                count += 1
    crit_b = crit_b[:count]

    return np.sort(np.concatenate((crit_a, crit_b)))


@njit(cache=True)
def _critical_b(A, B, P):
    v = B - A
    w = P - A
    vv = np.dot(v, v)
    if vv == 0:
        t = np.float64(0)
        return np.linalg.norm(w), t
    t = np.dot(v, w) / vv
    if t < 0:
        dist = np.linalg.norm(w)
        t = np.float64(0)
    elif t > 1:
        dist = np.linalg.norm(P - B)
        t = np.float64(1)
    else:
        dist = np.linalg.norm(t * v - w)
    return dist, t


@njit(cache=True)
def _critical_c(A, B, P1, P2):
    M = (P1 + P2) / 2
    AB = B - A
    MA = A - M
    PP = P2 - P1

    a = np.dot(AB, PP)
    b = np.dot(MA, PP)
    if a == 0:
        if np.abs(np.dot(AB, MA)) == np.linalg.norm(AB) * np.linalg.norm(MA):
            # M is on AB
            ret = np.linalg.norm(A - P1)
            t = np.float64(0)
        else:
            ret = NAN
            t = NAN
    else:
        t = -b / a
        if t < 0 or t > 1:
            ret = NAN
        else:
            ret = np.linalg.norm(A + AB * t - P1)
    return ret, t


@njit(cache=True)
def _free_boundaries(P, Q, eps):
    p, q = len(P), len(Q)
    BF = np.empty((p - 1, q, 2), dtype=np.float64)
    LF = np.empty((p, q - 1, 2), dtype=np.float64)

    for i in range(p - 1):
        for j in range(q):
            BF[i, j] = _free_interval(P[i], P[i + 1], Q[j], eps)
    for i in range(p):
        for j in range(q - 1):
            LF[i, j] = _free_interval(Q[j], Q[j + 1], P[i], eps)

    return BF, LF


@njit(cache=True)
def _reachable_boundaries(BF, LF, BR_out, LR_out):
    # Downmost boundary
    if BF[0, 0, 0] == 0:
        BR_out[0, 0] = BF[0, 0]
    else:
        BR_out[0, 0] = [NAN, NAN]
    for i in range(1, BR_out.shape[0]):
        if BR_out[i - 1, 0, 1] == 1 and BF[i, 0, 0] == 0:
            BR_out[i, 0] = BF[i, 0]
        else:
            BR_out[i, 0] = [NAN, NAN]

    # Leftmost boundary
    if LF[0, 0, 0] == 0:
        LR_out[0, 0] = LF[0, 0]
    else:
        LR_out[0, 0] = [NAN, NAN]
    for j in range(1, LR_out.shape[1]):
        if LR_out[0, j - 1, 1] == 1 and LF[0, j, 0] == 0:
            LR_out[0, j] = LF[0, j]
        else:
            LR_out[0, j] = [NAN, NAN]

    # Propagate
    for i in range(BR_out.shape[0]):
        for j in range(LR_out.shape[1]):
            prevB_start, _ = BR_out[i, j]
            B_start, B_end = BF[i, j + 1]
            prevL_start, _ = LR_out[i, j]
            L_start, L_end = LF[i + 1, j]

            if not np.isnan(prevL_start):
                BR_out[i, j + 1] = [B_start, B_end]
            elif prevB_start <= B_end:
                BR_out[i, j + 1] = [max(prevB_start, B_start), B_end]
            else:
                BR_out[i, j + 1] = [NAN, NAN]

            if not np.isnan(prevB_start):
                LR_out[i + 1, j] = [L_start, L_end]
            elif prevL_start <= L_end:
                LR_out[i + 1, j] = [max(prevL_start, L_start), L_end]
            else:
                LR_out[i + 1, j] = [NAN, NAN]

    return BR_out, LR_out


@njit(cache=True)
def _reachable_boundaries_1d(P, Q, eps):
    # Keep 1d array instead of 2d, and do free + reachable in one-shot.
    # Memory efficient, but cannot do backtracking.
    p, q = len(P), len(Q)
    B = np.empty((1, 2), dtype=np.float64)
    L = np.empty((q - 1, 2), dtype=np.float64)

    # Construct leftmost Ls
    prevL0_end = 1
    for j in range(q - 1):
        if prevL0_end == 1:
            start, end = _free_interval(Q[j], Q[j + 1], P[0], eps)
            if start == 0:
                L[j] = [start, end]
            else:
                L[j] = [NAN, NAN]
        else:
            L[j] = [NAN, NAN]
        _, prevL0_end = L[j]

    prevB0_end = 1
    for i in range(p - 1):
        # construct lowermost B
        if prevB0_end == 1:
            start, end = _free_interval(P[i], P[i + 1], Q[0], eps)
            if start == 0:
                B[0] = [start, end]
            else:
                B[0] = [NAN, NAN]
        else:
            B[0] = [NAN, NAN]
        _, prevB0_end = B[0]
        for j in range(q - 1):
            prevL_start, _ = L[j]
            prevB_start, _ = B[0]
            L_start, L_end = _free_interval(Q[j], Q[j + 1], P[i + 1], eps)
            B_start, B_end = _free_interval(P[i], P[i + 1], Q[j + 1], eps)

            if not np.isnan(prevB_start):
                L[j] = [L_start, L_end]
            elif prevL_start <= L_end:
                L[j] = [max(prevL_start, L_start), L_end]
            else:
                L[j] = [NAN, NAN]

            if not np.isnan(prevL_start):
                B[0] = [B_start, B_end]
            elif prevB_start <= B_end:
                B[0] = [max(prevB_start, B_start), B_end]
            else:
                B[0] = [NAN, NAN]

    return B, L


@njit(cache=True)
def _free_interval(A, B, P, eps):
    # resulting interval is always in [0, 1] or is [nan, nan].
    start = end = NAN

    pa = A - P
    pb = B - P
    if np.dot(pa, pa) <= eps**2:
        start = np.float64(0)
    if np.dot(pb, pb) <= eps**2:
        end = np.float64(1)

    ab = B - A
    a = np.dot(ab, ab)
    if a == 0:
        return [start, end]

    b = 2 * np.dot(ab, pa)
    c = np.dot(pa, pa) - eps**2
    Det = b**2 - 4 * a * c
    if Det >= 0:
        if np.isnan(start):  # pa.pa > eps ** 2
            start = (-b - Det**0.5) / 2 / a
        if np.isnan(end):  # pb.pb > eps ** 2
            end = (-b + Det**0.5) / 2 / a
    if start > 1 or end < 0:
        start = end = NAN
    return [start, end]



"""Integral Fréchet distance."""

import numpy as np
from numba import njit

from ._algorithms.ifd import _ifd_acm, _ifd_acm_1d, _ifd_owp
from .util import index2arclength

__all__ = [
    "ifd",
    "ifd_owp",
]


NAN = np.float64(np.nan)
EPSILON = np.finfo(np.float64).eps


@njit(cache=True)
def ifd(P, Q, delta, dist="euclidean"):
    r"""Integral Fréchet distance between two open polygonal curves.

    Let :math:`f, g: [0, 1] \to \Omega` be curves defined in a metric space
    :math:`\Omega`. Let :math:`\alpha, \beta: [0, 1] \to [0, 1]` be continuous
    non-decreasing surjections, and define :math:`\pi: [0, 1] \to [0, 1] \times
    [0, 1]` such that :math:`\pi(t) = \left(\alpha(t), \beta(t)\right)`.
    The integral Fréchet distance between :math:`f` and :math:`g` is defined as

    .. math::

        \inf_{\pi} \int_0^1
        dist\left(\pi(t)\right) \cdot
        \lVert \pi'(t) \rVert_1
        \mathrm{d}t,

    where :math:`dist\left(\pi(t)\right)` is a distance between
    :math:`f\left(\alpha(t)\right)` and :math:`g\left(\beta(t)\right)` and
    :math:`\lVert \cdot \rVert_1` is the Manhattan norm.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    delta : double
        Maximum length of edges between Steiner points.
        Refer to the Reference section for more information.
    dist : {'euclidean', 'squared_euclidean'}
        Type of :math:`dist`. Refer to the Notes section for more information.

    Returns
    -------
    double
        The integral Fréchet distance between *P* and *Q*, NaN if any vertice
        is empty or both vertices consist of a single point.

    See Also
    --------
    ifd_owp

    Notes
    -----
    This function implements the algorithm of Brankovic et al [1]_.

    The following functions are available for :math:`dist`:

    1. Euclidean distance
        .. math::

            dist\left(p, q\right) = \lVert p - q \rVert_2

        .. note::

            This distance is not implemented yet.

    2. Squared Euclidean distance
        .. math::

            dist\left(p, q\right) = \lVert p - q \rVert_2^2

    References
    ----------
    .. [1] Brankovic, M., et al. "(k, l)-Medians Clustering of Trajectories Using
       Continuous Dynamic Time Warping." Proceedings of the 28th International
       Conference on Advances in Geographic Information Systems. 2020.

    Examples
    --------
    >>> P, Q = [[0, 0], [0.5, 0], [1, 0]], [[0, 1], [1, 1]]
    >>> ifd(np.asarray(P), np.asarray(Q), 0.1, "squared_euclidean")
    2.0
    """
    B, L = _ifd_acm_1d(P.astype(np.float64), Q.astype(np.float64), delta, dist)
    if len(B) == 0 or len(L) == 0:
        ret = NAN
    else:
        ret = L[-1]
    return ret


@njit(cache=True)
def ifd_owp(P, Q, delta, dist="euclidean", param_type="arc-length"):
    """Integral Fréchet distance and its optimal warping path.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    delta : double
        Maximum length of edges between Steiner points. Refer to :func:`ifd`.
    dist : {'euclidean', 'squared_euclidean'}
        Type of :math:`dist`. Refer to :func:`ifd`.
    param_type : {'arc-length', 'vertex'}
        Parametrization of matching.

    Returns
    -------
    ifd : double
        The integral Fréchet distance between *P* and *Q*, NaN if any vertice
        is empty or both vertices consist of a single point.
    owp : ndarray
        Optimal warping path, empty if any vertice is empty or both vertices
        consist of a single point.

    Examples
    --------
    >>> from curvesimilarities.util import parameter_space
    >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
    >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
    >>> _, path = ifd_owp(P, Q, 0.1, "squared_euclidean")
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    >>> weight, p, q, _, _ = parameter_space(P, Q, 200, 100)
    >>> plt.pcolormesh(p, q, weight.T, cmap="gray")  # doctest: +SKIP
    >>> plt.plot(*path.T, "--")  # doctest: +SKIP
    """
    P, Q = P.astype(np.float64), Q.astype(np.float64)
    B, L = _ifd_acm(P, Q, delta, dist)
    if len(B) == 0 or len(L) == 0:
        ifd = NAN
    else:
        ifd = L[-1, -1]
    path = _ifd_owp(P, Q, B, L, delta, dist)[::-1]

    if param_type == "arc-length":
        path = np.stack(
            (
                index2arclength(P, path[:, 0].copy()),
                index2arclength(Q, path[:, 1].copy()),
            )
        ).T
    elif param_type == "vertex":
        pass
    else:
        raise ValueError("Unknown option for parametrization.")

    return ifd, path



"""Curve similarity measures."""

from .dtw import dtw, dtw_owp
from .frechet import dfd, dfd_idxs, fd
from .integfrechet import ifd, ifd_owp

__all__ = [
    "fd",
    "dfd",
    "dfd_idxs",
    "dtw",
    "dtw_owp",
    "ifd",
    "ifd_owp",
]



import numpy as np
from scipy.spatial.distance import cdist

from curvesimilarities import dtw


def test_dtw_degenerate():

    assert np.isnan(dtw(np.empty((0, 2)), np.empty((0, 2))))

    def check(P, Q):
        P, Q = np.asarray(P), np.asarray(Q)
        assert dtw(P, Q) == np.sum(cdist(P, Q))

    check([[0, 0]], [[0, 1]])
    check([[0, 0], [1, 0]], [[0, 1]])
    check([[0, 0]], [[0, 1], [1, 1]])


def test_dtw_duplicate(P_pts, Q_pts):
    # xfail because DTW does not sanitize the vertices.
    P_dup = np.repeat(P_pts, 2, axis=0)
    Q_dup = np.repeat(Q_pts, 2, axis=0)
    assert dtw(P_dup, Q_dup) != dtw(P_pts, Q_pts)



"""Utility functions."""

import numpy as np
from numba import njit
from scipy.spatial.distance import cdist

__all__ = [
    "parameter_space",
    "matching_pairs",
    "sample_polyline",
]


def parameter_space(P, Q, p_num, q_num):
    r"""Parameter space betwee two polylines.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    p_num, q_num : int
        Number of sample points in `P` and `Q`, respectively.

    Returns
    -------
    weight : ndarray
        A `p_num` by `q_num` array containing the distance between the points in
        `P` and `Q`.
    p_coord, q_coord : ndarray
        Axis coordinates for the parameter space.
    p_vert, q_vert : ndarray
        Coordinates for the vertices of polylines.

    Examples
    --------
    Curve space:

    .. plot::
        :context: close-figs

        >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
        >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
        >>> import matplotlib.pyplot as plt  # doctest: +SKIP
        >>> plt.plot(*P.T)  # doctest: +SKIP
        >>> plt.plot(*Q.T)  # doctest: +SKIP

    Parameter space with vertices as dashed lines:

    .. plot::
        :context: close-figs

        >>> weight, p, q, p_vert, q_vert = parameter_space(P, Q, 200, 100)
        >>> plt.pcolormesh(p, q, weight.T)  # doctest: +SKIP
        >>> plt.vlines(p_vert, 0, q[-1], "k", "--")  # doctest: +SKIP
        >>> plt.hlines(q_vert, 0, p[-1], "k", "--")  # doctest: +SKIP
    """
    p_vert = np.insert(np.cumsum(np.linalg.norm((np.diff(P, axis=0)), axis=-1)), 0, 0)
    p_coord = np.linspace(0, p_vert[-1], p_num)
    P_pts = sample_polyline(P, p_coord)
    q_vert = np.insert(np.cumsum(np.linalg.norm((np.diff(Q, axis=0)), axis=-1)), 0, 0)
    q_coord = np.linspace(0, q_vert[-1], q_num)
    Q_pts = sample_polyline(Q, q_coord)
    return cdist(P_pts, Q_pts), p_coord, q_coord, p_vert, q_vert


def matching_pairs(P, Q, path, sample_num):
    """Sample a continuous path in parameter space to matching pairs in curve space.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices in an
        :math:`n`-dimensional space.
    path : ndarray
        A :math:`N` by :math:`2` array of :math:`N` vertices of polyline in
        parameter space.
    sample_num : int
        Number of sample points to be uniformly taken from `path`.

    Returns
    -------
    ndarray
        A :math:`n` by :math:`2` by `sample_num` array of point pairs in curve space.

    Examples
    --------
    >>> from curvesimilarities import ifd_owp
    >>> from curvesimilarities.util import matching_pairs
    >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
    >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
    >>> _, path = ifd_owp(P, Q, 0.1, "squared_euclidean")
    >>> pairs = matching_pairs(P, Q, path, 50)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    >>> plt.plot(*P.T); plt.plot(*Q.T)  # doctest: +SKIP
    >>> plt.plot(*pairs, "--", color="gray")  # doctest: +SKIP
    """
    path_len = np.sum(np.linalg.norm(np.diff(path, axis=0), axis=-1))
    path_pts = sample_polyline(path, np.linspace(0, path_len, sample_num))
    P_pts = sample_polyline(P, path_pts[:, 0])
    Q_pts = sample_polyline(Q, path_pts[:, 1])
    return np.stack([P_pts, Q_pts]).transpose(2, 0, 1)


@njit(cache=True)
def sample_polyline(vert, param, param_type="arc-length"):
    """Sample points from a polyline.

    Parameters
    ----------
    vert : ndarray
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    param : array_like
        An 1-D array of :math:`q` parameters for sampled points, using arc-length
        parametrization.
    param_type : {'arc-length', 'vertex'}
        Parametrization type of *param*.

    Returns
    -------
    ndarray
        A :math:`q` by :math:`n` array of sampled points.
    """
    edges = (vert[1:] - vert[:-1]).astype(np.float64)
    edges_len = np.empty(len(edges), dtype=np.float64)
    for i in range(len(edges_len)):
        edges_len[i] = np.linalg.norm(edges[i])
    ret = np.empty((len(param), vert.shape[1]), dtype=np.float64)

    if param_type == "arc-length":
        len_cumsum = np.empty(len(vert), dtype=np.float64)
        len_cumsum[0] = 0
        for i in range(1, len(len_cumsum)):
            len_cumsum[i] = len_cumsum[i - 1] + edges_len[i - 1]
        pt_vert_idx = np.clip(np.searchsorted(len_cumsum, param) - 1, 0, None)
        for i in range(len(ret)):
            n = pt_vert_idx[i]
            t = param[i] - len_cumsum[n]
            if t == 0:
                ret[i] = vert[n]
            else:
                ret[i] = vert[n] + (edges[n] / edges_len[n]) * t
    elif param_type == "vertex":
        for i in range(len(ret)):
            n = int(param[i])
            t = param[i] - n
            if t == 0:
                ret[i] = vert[n]
            else:
                ret[i] = vert[n] + edges[n] * t
    else:
        raise ValueError("Unknown option for parametrization.")
    return ret


@njit(cache=True)
def index2arclength(curve, param):
    """Convert index based parameters of a curve to arc length based parameters.

    Parameters
    ----------
    curve : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices in an
        :math:`n`-dimensional space.
    param : ndarray
        Index based parameters.
    """
    orig_shape = param.shape
    param = param.reshape(-1)

    len_cumsum = np.empty(len(curve), dtype=np.float64)
    len_cumsum[0] = 0
    for i in range(1, len(curve)):
        len_cumsum[i] = len_cumsum[i - 1] + np.linalg.norm(curve[i] - curve[i - 1])

    ret = np.empty_like(param, dtype=np.float64)
    for i in range(len(param)):
        n = int(param[i])
        t = param[i] - n
        if t == 0:
            ret[i] = len_cumsum[n]
        else:
            ret[i] = len_cumsum[n] + t * (len_cumsum[n + 1] - len_cumsum[n])
    return ret.reshape(orig_shape)



import numpy as np
from numba import njit


@njit(cache=True)
def _dfd_ca(P, Q):
    # Eiter, T., & Mannila, H. (1994)
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    p, q = len(P), len(Q)
    ret = np.empty((p, q), dtype=np.float64)

    if p > 0 and q > 0:
        ret[0, 0] = np.linalg.norm(P[0] - Q[0])

    for i in range(1, p):
        ret[i, 0] = max(ret[i - 1, 0], np.linalg.norm(P[i] - Q[0]))

    for j in range(1, q):
        ret[0, j] = max(ret[0, j - 1], np.linalg.norm(P[0] - Q[j]))

    for i in range(1, p):
        for j in range(1, q):
            ret[i, j] = max(
                min(ret[i - 1, j], ret[i, j - 1], ret[i - 1, j - 1]),
                np.linalg.norm(P[i] - Q[j]),
            )

    return ret


@njit(cache=True)
def _dfd_ca_1d(P, Q):
    # Equivalent to _dfd_ca, but keep 1d array instead of 2d.
    # Memory efficient, but cannot do backtracking.
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    p, q = len(P), len(Q)
    ret = np.empty(q, dtype=np.float64)

    if p > 0 and q > 0:
        ret[0] = np.linalg.norm(P[0] - Q[0])

    for j in range(1, q):
        ret[j] = max(ret[j - 1], np.linalg.norm(P[0] - Q[j]))

    for i in range(1, p):
        left = ret[0]
        ret[0] = max(left, np.linalg.norm(P[i] - Q[0]))
        for j in range(1, q):
            diag = left
            left = ret[j]
            ret[j] = max(min(diag, left, ret[j - 1]), np.linalg.norm(P[i] - Q[j]))

    return ret


@njit(cache=True)
def _dfd_idxs(ca):
    p, q = ca.shape
    i, j = p - 1, q - 1

    while i > 0 or j > 0:
        current = ca[i, j]
        LEFT = np.inf if i == 0 else ca[i - 1, j]
        DOWN = np.inf if j == 0 else ca[i, j - 1]
        DIAG = np.inf if (i == 0 or j == 0) else ca[i - 1, j - 1]
        prev = min(LEFT, DOWN, DIAG)
        if current > prev:
            break
        elif current == LEFT:
            i -= 1
        elif current == DOWN:
            j -= 1
        else:
            i -= 1
            j -= 1
    return (i, j)



"""Dynamic time warping distance.

This module implements only the basic algorithm. If you need advanced features, use
dedicated packages such as `dtw-python
<https://pypi.org/project/dtw-python/>`_.
"""

import numpy as np
from numba import njit

from ._algorithms.dtw import _dtw_acm, _dtw_acm_1d, _dtw_owp

__all__ = [
    "dtw",
    "dtw_owp",
]


NAN = np.float64(np.nan)


@njit(cache=True)
def dtw(P, Q, dist="euclidean"):
    r"""Dynamic time warping distance between two ordered sets of points.

    Let :math:`\{P_0, P_1, ..., P_n\}` and :math:`\{Q_0, Q_1, ..., Q_m\}` be ordered
    sets of points in metric space. The dynamic time warping distance between
    two sets is defined as

    .. math::

        \min_{C} \sum_{(i, j) \in C} \lVert P_i - Q_j \rVert,

    where :math:`C` is a nondecreasing coupling over
    :math:`\{0, ..., n\} \times \{0, ..., m\}`, starting from :math:`(0, 0)` and
    ending with :math:`(n, m)`. :math:`\lVert \cdot \rVert` is the underlying
    metric.

    Parameters
    ----------
    P : ndarray
        A :math:`p` by :math:`n` array of :math:`p` points in an :math:`n`-dimensional
        space.
    Q : ndarray
        A :math:`q` by :math:`n` array of :math:`q` points in an :math:`n`-dimensional
        space.
    dist : {'euclidean', 'squared_euclidean'}
        Type of underlying metric. Refer to the Notes section for more information.

    Returns
    -------
    double
        The dynamic time warping distance between *P* and *Q*, NaN if any
        array of points is empty.

    See Also
    --------
    dtw_owp

    Notes
    -----
    This function implements the algorithm described by Senin [1]_.

    The following functions are available for :math:`\lVert \cdot \rVert`:

    1. Euclidean distance
        .. math::

            \lVert p - q \rVert = \lVert p - q \rVert_2

    2. Squared Euclidean distance
        .. math::

            \lVert p - q \rVert = \lVert p - q \rVert_2^2

    References
    ----------
    .. [1] Senin, P. (2008). Dynamic time warping algorithm review. Information
        and Computer Science Department University of Hawaii at Manoa Honolulu,
        USA, 855(1-23), 40.

    Examples
    --------
    >>> P = np.linspace([0, 0], [1, 0], 10)
    >>> Q = np.linspace([0, 1], [1, 1], 20)
    >>> dtw(P, Q)
    20.0...
    """
    acm = _dtw_acm_1d(P.astype(np.float64), Q.astype(np.float64), dist)
    if acm.size == 0:
        ret = NAN
    else:
        ret = acm[-1]
    return ret


@njit(cache=True)
def dtw_owp(P, Q, dist="euclidean"):
    """Dynamic time warping distance and its optimal warping path.

    Parameters
    ----------
    P : ndarray
        A :math:`p` by :math:`n` array of :math:`p` points in an :math:`n`-dimensional
        space.
    Q : ndarray
        A :math:`q` by :math:`n` array of :math:`q` points in an :math:`n`-dimensional
        space.
    dist : {'euclidean', 'squared_euclidean'}
        Type of underlying metric. Refer to :func:`dtw`.

    Returns
    -------
    dtw : double
        The dynamic time warping distance between *P* and *Q*, NaN if any
        array of points is empty.
    owp : ndarray
        Indices of *P* and *Q* for optimal warping path, empty if any array of points
        empty.

    Examples
    --------
    >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
    >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
    >>> _, owp = dtw_owp(P, Q)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    >>> x, y = np.meshgrid(np.arange(len(P)), np.arange(len(Q)))
    >>> plt.plot(*np.vstack([x.ravel(), y.ravel()]), "x")  # doctest: +SKIP
    >>> plt.plot(*owp.T, "o")  # doctest: +SKIP
    >>> plt.axis("equal")  # doctest: +SKIP
    """
    acm = _dtw_acm(P.astype(np.float64), Q.astype(np.float64), dist)
    if acm.size == 0:
        ret = NAN, np.empty((0, 2), dtype=np.int_)
    else:
        ret = acm[-1, -1], _dtw_owp(acm)
    return ret



"""Continuous and discrete Fréchet distances."""

import numpy as np
from numba import njit

from ._algorithms.dfd import _dfd_ca, _dfd_ca_1d, _dfd_idxs
from ._algorithms.fd import _fd, _reachable_boundaries_1d
from ._algorithms.lcfm import _computeLCFM, _significant_events
from .util import index2arclength

__all__ = [
    "fd",
    "decision_problem",
    "significant_events",
    "fd_matching",
    "dfd",
    "dfd_idxs",
]


EPSILON = np.finfo(np.float64).eps
NAN = np.float64(np.nan)


@njit(cache=True)
def fd(P, Q, rel_tol=0.0, abs_tol=float(EPSILON)):
    r"""(Continuous) Fréchet distance between two open polylines.

    Let :math:`f: [0, 1] \to \Omega` and :math:`g: [0, 1] \to \Omega` be curves
    in a metric space :math:`\Omega`. The Fréchet distance between
    :math:`f` and :math:`g` is defined as

    .. math::

        \inf_{\alpha, \beta} \max_{t \in [0, 1]}
        \lVert f(\alpha(t)) - g(\beta(t)) \rVert,

    where :math:`\alpha, \beta: [0, 1] \to [0, 1]` are continuous non-decreasing
    surjections and :math:`\lVert \cdot \rVert` is the underlying metric.
    The Euclidean metric is used in this function.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    rel_tol, abs_tol : double
        Relative and absolute tolerances for parametric search of the Fréchet distance.

    Returns
    -------
    dist : double
        The (continuous) Fréchet distance between *P* and *Q*, NaN if any vertex
        is empty.

    See Also
    --------
    decision_problem
    significant_events
    fd_matching

    Notes
    -----
    This function implements Alt and Godau's algorithm [1]_.

    The parametric search for the Fréchet distance terminates if the upper boundary
    ``a`` and the lower boundary ``b`` satisfy: ``a - b <= max(rel_tol * a, abs_tol)``.

    References
    ----------
    .. [1] Alt, H., & Godau, M. (1995). Computing the Fréchet distance between
       two polygonal curves. International Journal of Computational Geometry &
       Applications, 5(01n02), 75-91.

    Examples
    --------
    >>> P, Q = [[0, 0], [0.5, 0], [1, 0]], [[0, 1], [1, 1]]
    >>> fd(np.asarray(P), np.asarray(Q))
    1.0...
    """
    return _fd(P.astype(np.float64), Q.astype(np.float64), rel_tol, abs_tol)


@njit(cache=True)
def decision_problem(P, Q, epsilon):
    """Decision problem of the (continuous) Fréchet distance.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    epsilon : double
        Minimum distance to be checked.

    Returns
    -------
    bool
        True if *epsilon* is greater than or equal to the Fréchet distance between
        *P* and *Q*, false otherwise.
    """
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    B, L = _reachable_boundaries_1d(P.astype(np.float64), Q.astype(np.float64), epsilon)
    if B[-1, 1] == 1 or L[-1, 1] == 1:
        ret = True
    else:
        ret = False
    return ret


@njit(cache=True)
def significant_events(
    P,
    Q,
    param_type="arc-length",
    rel_tol=0.0,
    abs_tol=float(EPSILON),
    event_rel_tol=0.0,
    event_abs_tol=float(EPSILON),
):
    """Return significant events of the (continuous) Fréchet distance [1]_ .

    A significant event is a matching which determines the Fréchet distance.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    param_type : {'arc-length', 'vertex'}
        Parametrization of matching.
    rel_tol, abs_tol : double
        Relative and absolute tolerances for parametric search of the feasible distance.
    event_rel_tol, event_abs_tol : double
        Relative and absolute tolerances to determine realizing events.

    Returns
    -------
    events : ndarray
        :math:`N` significant events in a :math:`(N, 2, 2)`-shaped array of parameters.
        The second axis is the starting and ending points of the event.
        The last axis is the parameters of *P* and *Q*.
    errors : ndarray
        Difference between the Fréchet distance and the values of the events.

    Notes
    -----
    This function implements Buchin et al.'s algorithm [1]_, except that backtracking is
    used to extract significant events. Plus, the "type-A" events are included.

    The parametric search for the feasible distance terminates if the upper boundary
    ``a`` and the lower boundary ``b`` satisfy: ``a - b <= max(rel_tol * a, abs_tol)``.

    An event is considered to be realizing if its value `e` and the realizing distnace
    `d` satisfy: ``e - d <= max(rel_tol * d, abs_tol)``.

    References
    ----------
    .. [1] Buchin, K., et al. "Locally correct Fréchet matchings."
       Computational Geometry 76 (2019): 1-18.

    Examples
    --------
    >>> from curvesimilarities.frechet import significant_events
    >>> from curvesimilarities.util import parameter_space
    >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
    >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
    >>> events, _ = significant_events(P, Q)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    >>> weight, p, q, _, _ = parameter_space(P, Q, 200, 100)
    >>> plt.pcolormesh(p, q, weight.T < fd(P, Q), cmap="gray")  # doctest: +SKIP
    >>> plt.plot(*events.transpose(2, 1, 0), "o-")  # doctest: +SKIP
    """
    P, Q = P.astype(np.float64), Q.astype(np.float64)
    p, q = len(P), len(Q)

    events = np.empty((p * q, 2, 2), dtype=np.float64)
    errors = np.empty(p * q, dtype=np.float64)
    count = 0
    if p == 0 or q == 0:
        return events, errors

    A0 = np.linalg.norm(P[0] - Q[0])
    A1 = np.linalg.norm(P[-1] - Q[-1])
    if p <= 2 and q <= 2:
        d = max(A0, A1)
    else:
        eps, BE, LE, BE_val, LE_val, _, _ = _significant_events(
            P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol
        )
        d = max(eps, A0, A1)
        if abs(d - eps) <= max(event_rel_tol * max(abs(eps), abs(d)), event_abs_tol):
            # BE and LE are significant events (if exist).
            for i in range(len(BE)):
                it, j0, j1 = BE[i]
                events[count, 0] = [it, j0]
                events[count, 1] = [it, j1]
                errors[count] = abs(d - BE_val[i])
                count += 1
            for j in range(len(LE)):
                i0, i1, jt = LE[j]
                events[count, 0] = [i0, jt]
                events[count, 1] = [i1, jt]
                errors[count] = abs(d - LE_val[i])
                count += 1
    if abs(d - A0) <= max(event_rel_tol * max(abs(A0), abs(d)), event_abs_tol):
        # A0 is a significant event.
        events[count, 0] = [0, 0]
        events[count, 1] = [0, 0]
        errors[count] = abs(d - A0)
        count += 1
    if abs(d - A1) <= max(event_rel_tol * max(abs(A1), abs(d)), event_abs_tol):
        # A1 is a significant event.
        events[count, 0] = [p - 1, q - 1]
        events[count, 1] = [p - 1, q - 1]
        errors[count] = abs(d - A1)
        count += 1
    events = events[:count]
    errors = errors[:count]

    if param_type == "arc-length":
        events = np.stack(
            (
                index2arclength(P, events[:, :, 0].copy()),
                index2arclength(Q, events[:, :, 1].copy()),
            )
        ).transpose(1, 2, 0)
    elif param_type == "vertex":
        pass
    else:
        raise ValueError("Unknown option for parametrization.")

    return events, errors


@njit(cache=True)
def fd_matching(
    P,
    Q,
    param_type="arc-length",
    rel_tol=0.0,
    abs_tol=float(EPSILON),
    event_rel_tol=0.0,
    event_abs_tol=float(EPSILON),
):
    """Locally correct Fréchet matching [1]_.

    A locally correct Fréchet matching is a Fréchet matching between two curves whose
    any sub-matching is still a Fréchet matching.

    Parameters
    ----------
    P : array_like
        A :math:`p` by :math:`n` array of :math:`p` vertices of a polyline in an
        :math:`n`-dimensional space.
    Q : array_like
        A :math:`q` by :math:`n` array of :math:`q` vertices of a polyline in an
        :math:`n`-dimensional space.
    param_type : {'arc-length', 'vertex'}
        Parametrization of matching.
    rel_tol, abs_tol : double
        Relative and absolute tolerances for parametric search of the Fréchet distance.
    event_rel_tol, event_abs_tol : double
        Relative and absolute tolerances to determine realizing events.

    Returns
    -------
    dist : double
        The (continuous) Fréchet distance between *P* and *Q*, NaN if any vertex
        is empty.
    matching : ndarray
        Vertices of a locally correct Fréchet matching in parameter space.

    Notes
    -----
    This function implements Buchin et al.'s algorithm [1]_, except that backtracking is
    used to extract significant events.

    The parametric search for the feasible distance terminates if the upper boundary
    ``a`` and the lower boundary ``b`` satisfy: ``a - b <= max(rel_tol * a, abs_tol)``.

    An event is considered to be realizing if its value `e` and the realizing distnace
    `d` satisfy: ``e - d <= max(rel_tol * d, abs_tol)``.

    References
    ----------
    .. [1] Buchin, Kevin, et al. "Locally correct Fréchet matchings."
       Computational Geometry 76 (2019): 1-18.

    Examples
    --------
    >>> from curvesimilarities.frechet import fd_matching
    >>> from curvesimilarities.util import parameter_space
    >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
    >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
    >>> dist, path = fd_matching(P, Q)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    >>> weight, p, q, _, _ = parameter_space(P, Q, 200, 100)
    >>> plt.pcolormesh(p, q, weight.T < dist, cmap="gray")  # doctest: +SKIP
    >>> plt.plot(*path.T, "--")  # doctest: +SKIP
    """
    P, Q = P.astype(np.float64), Q.astype(np.float64)
    eps, matching = _computeLCFM(P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol)
    if not np.isnan(eps):
        dist = max(eps, np.linalg.norm(P[0] - Q[0]), np.linalg.norm(P[-1] - Q[-1]))
    else:
        dist = max(np.linalg.norm(P[0] - Q[0]), np.linalg.norm(P[-1] - Q[-1]))
    if param_type == "arc-length":
        matching = np.stack(
            (
                index2arclength(P, matching[:, 0].copy()),
                index2arclength(Q, matching[:, 1].copy()),
            )
        ).T
    elif param_type == "vertex":
        pass
    else:
        raise ValueError("Unknown option for parametrization.")
    return dist, matching


@njit(cache=True)
def dfd(P, Q):
    r"""Discrete Fréchet distance between two two ordered sets of points.

    Let :math:`\{P_0, P_1, ..., P_n\}` and :math:`\{Q_0, Q_1, ..., Q_m\}` be ordered
    sets of points in metric space. The discrete Fréchet distance between two sets is
    defined as

    .. math::

        \min_{C} \max_{(i, j) \in C} \lVert P_i - Q_j \rVert,

    where :math:`C` is a nondecreasing coupling over
    :math:`\{0, ..., n\} \times \{0, ..., m\}`, starting from :math:`(0, 0)` and
    ending with :math:`(n, m)`. :math:`\lVert \cdot \rVert` is the underlying
    metric, which is the Euclidean metric in this function.

    Parameters
    ----------
    P : ndarray
        A :math:`p` by :math:`n` array of :math:`p` points in an :math:`n`-dimensional
        space.
    Q : ndarray
        A :math:`q` by :math:`n` array of :math:`q` points in an :math:`n`-dimensional
        space.

    Returns
    -------
    dist : double
        The discrete Fréchet distance between *P* and *Q*, NaN if any array of points
        is empty.

    Notes
    -----
    This function implements Eiter and Mannila's algorithm [1]_.

    References
    ----------
    .. [1] Eiter, T., & Mannila, H. (1994). Computing discrete Fréchet distance.

    Examples
    --------
    >>> P, Q = [[0, 0], [1, 1], [2, 0]], [[0, 1], [2, -4]]
    >>> dfd(np.asarray(P), np.asarray(Q))
    4.0
    """
    ca = _dfd_ca_1d(P.astype(np.float64), Q.astype(np.float64))
    if ca.size == 0:
        ret = NAN
    else:
        ret = ca[-1]
    return ret


@njit(cache=True)
def dfd_idxs(P, Q):
    """Discrete Fréchet distance and its indices in curve space.

    Parameters
    ----------
    P : ndarray
        An :math:`p` by :math:`n` array of :math:`p` points in an
        :math:`n`-dimensional space.
    Q : ndarray
        An :math:`q` by :math:`n` array of :math:`q` points in an
        :math:`n`-dimensional space.

    Returns
    -------
    d : double
        The discrete Fréchet distance between *P* and *Q*, NaN if any array of points
        is empty.
    index_1 : int
        Index of point contributing to discrete Fréchet distance in *P*.
    index_2 : int
        Index of point contributing to discrete Fréchet distance in *Q*.

    Examples
    --------
    >>> P = np.array([[0, 0], [2, 2], [4, 2], [4, 4], [2, 1], [5, 1], [7, 2]])
    >>> Q = np.array([[2, 0], [1, 3], [5, 3], [5, 2], [7, 3]])
    >>> from curvesimilarities.util import sample_polyline
    >>> P_len = np.sum(np.linalg.norm(np.diff(P, axis=0), axis=-1))
    >>> P_pts = sample_polyline(P, np.linspace(0, P_len, 30))
    >>> Q_len = np.sum(np.linalg.norm(np.diff(Q, axis=0), axis=-1))
    >>> Q_pts = sample_polyline(Q, np.linspace(0, Q_len, 30))
    >>> _, idx0, idx1 = dfd_idxs(P_pts, Q_pts)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    >>> plt.plot(*P_pts.T, "x"); plt.plot(*Q_pts.T, "x")  # doctest: +SKIP
    >>> plt.plot(*np.array([P_pts[idx0], Q_pts[idx1]]).T, "--")  # doctest: +SKIP
    """
    ca = _dfd_ca(P.astype(np.float64), Q.astype(np.float64))
    if ca.size == 0:
        ret = NAN, -1, -1
    else:
        index_1, index_2 = _dfd_idxs(ca)
        ret = ca[-1, -1], int(index_1), int(index_2)
    return ret



"""Locally corret frechet matching."""

import numpy as np
from numba import njit

from .fd import _critical_b, _critical_c, _free_boundaries, _reachable_boundaries

EPSILON = np.finfo(np.float64).eps
NAN = np.float64(np.nan)
INF = np.float64(np.inf)


@njit(cache=True)
def _computeLCFM(P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol):
    if len(P.shape) != 2:
        raise ValueError("P must be a 2-dimensional array.")
    if len(Q.shape) != 2:
        raise ValueError("Q must be a 2-dimensional array.")
    if P.shape[1] != Q.shape[1]:
        raise ValueError("P and Q must have the same number of columns.")

    p, q = len(P), len(Q)
    eps = NAN
    if p == 0 or q == 0:
        matching = np.empty((0, 2), dtype=np.float64)
    elif p == 1 and q == 1:
        matching = np.array([[0, 0]], dtype=np.float64)
    elif p == 1:
        if q > 2:
            dists = np.empty(q - 2, dtype=np.float64)
            for i in range(0, q - 2):
                dists[i] = np.linalg.norm(P[0] - Q[i + 1])
            eps = np.max(dists)
        matching = np.array([[0, 0], [0, q - 1]], dtype=np.float64)
    elif q == 1:
        if p > 2:
            dists = np.empty(p - 2, dtype=np.float64)
            for i in range(0, p - 2):
                dists[i] = np.linalg.norm(P[i + 1] - Q[0])
            eps = np.max(dists)
        matching = np.array([[0, 0], [p - 1, 0]], dtype=np.float64)
    elif p == 2 and q == 2:
        matching = np.array([[0, 0], [1, 1]], dtype=np.float64)
    else:
        eps, e_r, e_r_horiz = _e_r(P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol)

        i0, j0 = e_r[0]
        P1 = _idx2subcurve(P, 0, i0)
        Q1 = _idx2subcurve(Q, 0, j0)
        i1, j1 = e_r[-1]
        P2 = _idx2subcurve(P, i1, len(P) - 1)
        Q2 = _idx2subcurve(Q, j1, len(Q) - 1)

        _, mu_1 = _computeLCFM(P1, Q1, rel_tol, abs_tol, event_abs_tol, event_rel_tol)
        _, mu_2 = _computeLCFM(P2, Q2, rel_tol, abs_tol, event_abs_tol, event_rel_tol)
        # Scale mu s.t. indices are w.r.t full curves instead of subcurves
        if e_r_horiz:
            J, T = int(j0), j0 - int(j0)
            thres = len(Q1) - 2
            for idx in range(len(mu_1)):
                _, j = mu_1[idx]
                if j > thres:
                    t = j - J
                    mu_1[idx, 1] = J + t * T
            mu_2[:, 0] += i1
            for idx in range(len(mu_2)):
                _, j = mu_2[idx]
                if j < 1:
                    mu_2[idx, 1] = J + T + j * (1 - T)
                else:
                    mu_2[idx, 1] += J
        else:
            I, T = int(i0), i0 - int(i0)
            thres = len(P1) - 2
            for idx in range(len(mu_1)):
                i, _ = mu_1[idx]
                if i > thres:
                    t = i - I
                    mu_1[idx, 0] = I + t * T
            mu_2[:, 1] += j1
            for idx in range(len(mu_2)):
                i, _ = mu_2[idx]
                if i < 1:
                    mu_2[idx, 0] = I + T + i * (1 - T)
                else:
                    mu_2[idx, 0] += I

        matching = np.concatenate((mu_1[:-1], e_r, mu_2[1:]))
    return eps, matching


@njit(cache=True)
def _significant_events(P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol):
    # STEP 1: compute feasible epsilon and its reachable boundaries.
    p, q = len(P), len(Q)

    crit = np.empty(p * (q - 1) + (p - 1) * q, dtype=np.float64)
    count = 0
    for i in range(p - 1):
        for j in range(q):
            crit[count], _ = _critical_b(P[i], P[i + 1], Q[j])
            count += 1
    for i in range(p):
        for j in range(q - 1):
            crit[count], _ = _critical_b(Q[j], Q[j + 1], P[i])
            count += 1
    crit = np.sort(crit[:count])
    # Compute minimum feasible epsilons: 1. Binary search
    start, end = 0, len(crit)
    BF, LF, BR, LR = _feasible_boundaries(P, Q, crit[start])
    if BR[-1, -1, 1] == 1 or LR[-1, -1, 1] == 1:
        end = start
    while end - start > 1:
        mid = (start + end) // 2
        BF, LF, BR, LR = _feasible_boundaries(P, Q, crit[mid])
        if BR[-1, -1, 1] == 1 or LR[-1, -1, 1] == 1:
            end = mid
        else:
            start = mid
    # Compute minimum feasible epsilons: 2. Parametric search
    e1, e2 = crit[start], crit[end]
    while e2 - e1 > max(rel_tol * e2, abs_tol):
        mid = (e1 + e2) / 2
        if (mid - e1 < EPSILON) or (e2 - mid < EPSILON):
            break
        BF, LF, BR, LR = _feasible_boundaries(P, Q, mid)
        if BR[-1, -1, 1] == 1 or LR[-1, -1, 1] == 1:
            e2 = mid
        else:
            e1 = mid
    eps = e2
    if not (BR[-1, -1, 1] == 1 or LR[-1, -1, 1] == 1):  # compute one more last time
        BF, LF, BR, LR = _feasible_boundaries(P, Q, eps)

    # STEP 2. Compute passable boundaries by backtracking.
    BP, LP = _passable_boundaries(BR, LR, BR, LR)

    # STEP 3. Determine realizing set with passable boundaries.
    # By using passable boundaries instead of reachable boundaries, insignificant events
    # are automatically excluded.
    BE, LE, BE_val, LE_val, BE_err, LE_err = _realizing_set(
        P, Q, eps, BF, LF, BP, LP, event_rel_tol, event_abs_tol
    )

    return eps, BE, LE, BE_val, LE_val, BE_err, LE_err


@njit(cache=True)
def _e_r(P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol):
    eps, BE, LE, _, _, BE_err, LE_err = _significant_events(
        P, Q, rel_tol, abs_tol, event_rel_tol, event_abs_tol
    )

    # Select the most significant event.
    if len(LE) == 0:
        e_r_horiz = False
        B_minerr_idx = np.argmin(BE_err)
    elif len(BE) == 0:
        e_r_horiz = True
        L_minerr_idx = np.argmin(LE_err)
    else:
        B_minerr_idx = np.argmin(BE_err)
        L_minerr_idx = np.argmin(LE_err)
        if BE_err[B_minerr_idx] < LE_err[L_minerr_idx]:
            e_r_horiz = False
        else:
            e_r_horiz = True

    if e_r_horiz:
        i0, i1, jt = LE[L_minerr_idx]
        if i0 == i1:
            e_r = np.array([[i0, jt]], dtype=np.float64)
        else:
            e_r = np.array([[i0, jt], [i1, jt]], dtype=np.float64)
    else:
        it, j0, j1 = BE[B_minerr_idx]
        if j0 == j1:
            e_r = np.array([[it, j0]], dtype=np.float64)
        else:
            e_r = np.array([[it, j0], [it, j1]], dtype=np.float64)

    return eps, e_r, e_r_horiz


@njit(cache=True)
def _feasible_boundaries(P, Q, eps):
    BF, LF = _free_boundaries(P, Q, eps)
    BF[0, 0] = [0, 1]
    LF[0, 0] = [0, 1]
    BF[-1, -1] = [0, 1]
    LF[-1, -1] = [0, 1]
    BR, LR = _reachable_boundaries(BF, LF, np.empty_like(BF), np.empty_like(LF))
    return BF, LF, BR, LR


@njit(cache=True)
def _passable_boundaries(BR, LR, BP_out, LP_out):
    # Uppermost boundary
    if BR[-1, -1, 1] == 1:
        BP_out[-1, -1] = BR[-1, -1]
    else:
        BP_out[-1, -1] = [NAN, NAN]
    for i in range(1, BP_out.shape[0]):
        if BP_out[-i, -1, 0] == 0 and BR[-i - 1, -1, 1] == 1:
            BP_out[-i - 1, -1] = BR[-i - 1, -1]
        else:
            BP_out[-i - 1, -1] = [NAN, NAN]

    # Rightmost boundary
    if LR[-1, -1, 1] == 1:
        LP_out[-1, -1] = LR[-1, -1]
    else:
        LP_out[-1, -1] = [NAN, NAN]
    for j in range(1, LP_out.shape[1]):
        if LP_out[-1, -j, 0] == 0 and LR[-1, -j - 1, 1] == 1:
            LP_out[-1, -j - 1] = LR[-1, -j - 1]
        else:
            LP_out[-1, -j - 1] = [NAN, NAN]

    for i in range(BP_out.shape[0]):
        for j in range(LP_out.shape[1]):
            _, prevB_end = BP_out[-i - 1, -j - 1]
            B_start, B_end = BR[-i - 1, -j - 2]
            _, prevL_end = LP_out[-i - 1, -j - 1]
            L_start, L_end = LR[-i - 2, -j - 1]

            if not np.isnan(prevL_end):
                BP_out[-i - 1, -j - 2] = [B_start, B_end]
            elif prevB_end >= B_start:
                BP_out[-i - 1, -j - 2] = [B_start, min(prevB_end, B_end)]
            else:
                BP_out[-i - 1, -j - 2] = [NAN, NAN]

            if not np.isnan(prevB_end):
                LP_out[-i - 2, -j - 1] = [L_start, L_end]
            elif prevL_end >= L_start:
                LP_out[-i - 2, -j - 1] = [L_start, min(prevL_end, L_end)]
            else:
                LP_out[-i - 2, -j - 1] = [NAN, NAN]

    return BP_out, LP_out


@njit(cache=True)
def _realizing_set(P, Q, eps, BF, LF, BR, LR, rel_tol, abs_tol):
    # Each row of LE = i0, i1, j + t
    LE = np.empty((LR.shape[0] * LR.shape[1], 3), dtype=np.float64)
    LE_val = np.empty(LR.shape[0] * LR.shape[1], dtype=np.float64)
    LE_err = np.empty(LR.shape[0] * LR.shape[1], dtype=np.float64)
    count = 0
    # Fallback variables
    fb_LE = np.empty((1, 3), dtype=np.float64)
    fb_LE_val = np.empty(1, dtype=np.float64)
    fb_LE_err = INF
    for j in range(LR.shape[1]):
        g_star = 0
        for i in range(1, LR.shape[0] - 1):  # skip event ending on rightmost boundary
            if not np.isnan(BR[i - 1, j, 0]) or LF[g_star, j, 0] <= LF[i, j, 0]:
                g_star = i
                d, t = _critical_b(Q[j], Q[j + 1], P[i])
            else:
                d, t = _critical_c(Q[j], Q[j + 1], P[g_star], P[i])
            if t == 1 and j < LR.shape[1] - 1 and LR[i, j + 1, 0] == 0:
                # Let upper cell deal with this event to prevent duplication.
                continue

            err = abs(eps - d)
            if j < LR.shape[1] - 1 and LR[i, j, 0] == 1 and LR[i, j + 1, 0] == 0:
                is_singleton = False
            elif j > 0 and LR[i, j, 1] == 0 and LR[i, j - 1, 1] == 1:
                is_singleton = False
            elif LR[i, j, 0] == LR[i, j, 1]:
                is_singleton = True
            else:  # Singleton detection may have failed due to floating point error
                is_realizing = err <= max(rel_tol * max(abs(eps), abs(d)), abs_tol)
                is_singleton = is_realizing & ~np.isnan(LR[i, j, 0])

            if is_singleton:
                LE[count] = [g_star, i, j + t]
                LE_val[count] = d
                LE_err[count] = err
                count += 1
                g_star = i

            # Update fallback
            if err < fb_LE_err:
                fb_LE[0, :] = [g_star, i, j + t]
                fb_LE_val[0] = d
                fb_LE_err = err

    LE = LE[:count]
    LE_val = LE_val[:count]
    LE_err = LE_err[:count]

    # Each row of BE = i + t, j0, j1
    BE = np.empty((BR.shape[0] * BR.shape[1], 3), dtype=np.float64)
    BE_val = np.empty(LR.shape[0] * LR.shape[1], dtype=np.float64)
    BE_err = np.empty(BR.shape[0] * BR.shape[1], dtype=np.float64)
    count = 0
    # Fallback variables
    fb_BE = np.empty((1, 3), dtype=np.float64)
    fb_BE_val = np.empty(1, dtype=np.float64)
    fb_BE_err = INF
    for i in range(BR.shape[0]):
        g_star = 0
        for j in range(1, BR.shape[1] - 1):  # skip event ending on uppermost boundary
            if not np.isnan(LR[i, j - 1, 0]) or BF[i, g_star, 0] <= BF[i, j, 0]:
                g_star = j
                d, t = _critical_b(P[i], P[i + 1], Q[j])
            else:
                d, t = _critical_c(P[i], P[i + 1], Q[g_star], Q[j])
            if t == 1 and i < BR.shape[0] - 1 and BR[i + 1, j, 0] == 0:
                # Let rhs cell deal with this event to prevent duplication.
                continue

            err = abs(eps - d)
            if i < BR.shape[0] - 1 and BR[i, j, 0] == 1 and BR[i + 1, j, 0] == 0:
                is_singleton = False
            elif i > 0 and BR[i, j, 1] == 0 and BR[i - 1, j, 1] == 1:
                is_singleton = False
            elif BR[i, j, 0] == BR[i, j, 1]:
                is_singleton = True
            else:  # Singleton detection may have failed due to floating point error
                is_realizing = err <= max(rel_tol * max(abs(eps), abs(d)), abs_tol)
                is_singleton = is_realizing & ~np.isnan(BR[i, j, 0])

            if is_singleton:
                BE[count] = [i + t, g_star, j]
                BE_val[count] = d
                BE_err[count] = err
                count += 1
                g_star = j

            # Update fallback
            if err < fb_BE_err:
                fb_BE[0, :] = [g_star, i, j + t]
                fb_BE_val[0] = d
                fb_BE_err = err

    BE = BE[:count]
    BE_val = BE_val[:count]
    BE_err = BE_err[:count]

    if len(BE) == 0 and len(LE) == 0:
        # Realizing event not detected because of too harsh tolerances.
        # Since at least one realizing event must exist, we regard the event whose value
        # is the closest to epsilon as the only realizing event.
        if fb_BE_err < fb_LE_err:
            BE = fb_BE
            BE_val = fb_BE_val
            BE_err = np.array((fb_BE_err,), dtype=np.float64)
        else:
            LE = fb_LE
            LE_val = fb_LE_val
            LE_err = np.array((fb_LE_err,), dtype=np.float64)

    return BE, LE, BE_val, LE_val, BE_err, LE_err


@njit(cache=True)
def _idx2subcurve(curve, idx0, idx1):
    i0, t0 = int(np.ceil(idx0)) - 1, idx0 - int(idx0)
    if t0 == 0:
        start = curve[i0 + 1]
    else:
        start = curve[i0] + t0 * (curve[i0 + 1] - curve[i0])

    if idx0 == idx1:
        return start.reshape(1, -1)

    i1, t1 = int(np.ceil(idx1)) - 1, idx1 - int(idx1)
    if t1 == 0:
        end = curve[i1 + 1]
    else:
        end = curve[i1] + t1 * (curve[i1 + 1] - curve[i1])
    return np.concatenate(
        (start.reshape(1, -1), curve[int(idx0) + 1 : i1 + 1], end.reshape(1, -1))
    )



import numpy as np
from scipy.spatial.distance import cdist

from curvesimilarities.frechet import decision_problem, dfd, dfd_idxs, fd, fd_matching


def test_fd_degenerate():

    assert np.isnan(fd(np.empty((0, 2)), np.empty((0, 2))))

    def check(P, Q):
        assert fd(np.asarray(P), np.asarray(Q)) == np.max(cdist(P, Q))

    check([[0, 0]], [[0, 1]])
    check([[0, 0], [1, 0]], [[0, 1]])
    check([[0, 0]], [[0, 1], [1, 1]])


def test_fd_duplicate(P_vert, Q_vert):
    P_dup = np.repeat(P_vert, 2, axis=0)
    Q_dup = np.repeat(Q_vert, 2, axis=0)
    assert fd(P_dup, Q_dup) == fd(P_vert, Q_vert)


def test_fd_decision_problem(P_vert, Q_vert):
    dist = fd(P_vert, Q_vert)
    assert not decision_problem(P_vert, Q_vert, dist / 2)
    assert decision_problem(P_vert, Q_vert, dist)


def test_fd_matching(P_vert, Q_vert):
    assert fd(P_vert[:1], Q_vert[:1]) == fd_matching(P_vert[:1], Q_vert[:1])[0]
    assert fd(P_vert[:1], Q_vert[:2]) == fd_matching(P_vert[:1], Q_vert[:2])[0]
    assert fd(P_vert[:2], Q_vert[:1]) == fd_matching(P_vert[:2], Q_vert[:1])[0]
    assert fd(P_vert[:1], Q_vert[:3]) == fd_matching(P_vert[:1], Q_vert[:3])[0]
    assert fd(P_vert[:3], Q_vert[:1]) == fd_matching(P_vert[:3], Q_vert[:1])[0]
    assert fd(P_vert[:2], Q_vert[:2]) == fd_matching(P_vert[:2], Q_vert[:2])[0]
    assert fd(P_vert, Q_vert) == fd_matching(P_vert, Q_vert)[0]


def test_dfd_degenerate():

    assert np.isnan(dfd(np.empty((0, 2)), np.empty((0, 2))))

    def check(P, Q):
        P, Q = np.asarray(P), np.asarray(Q)
        assert dfd(P, Q) == np.max(cdist(P, Q))

    check([[0, 0]], [[0, 1]])
    check([[0, 0], [1, 0]], [[0, 1]])
    check([[0, 0]], [[0, 1], [1, 1]])


def test_dfd_duplicate(P_pts, Q_pts):
    P_dup = np.repeat(P_pts, 2, axis=0)
    Q_dup = np.repeat(Q_pts, 2, axis=0)
    assert dfd(P_dup, Q_dup) == dfd(P_pts, Q_pts)


def test_dfd_idxs(P_pts, Q_pts):
    dist = cdist(P_pts, Q_pts)
    d, i, j = dfd_idxs(P_pts, Q_pts)
    assert d == dist[i, j]
    assert d == dfd(P_pts, Q_pts)
